#include "helper.h"
#include "alloc_page.h"
#include "delay.h"
#include "desc.h"
#include "isr.h"
#include "libcflat.h"
#include "msr.h"
#include "processor.h"
#include "smp.h"
#include "svm.h"
#include "vm.h"

void vmcb_set_seg(struct vmcb_seg *seg, u16 selector, u64 base, u32 limit,
                  u32 attr) {
  seg->selector = selector;
  seg->attrib = attr;
  seg->limit = limit;
  seg->base = base;
}

inline void vmmcall(void) { asm volatile("vmmcall" : : : "memory"); }

u8 *io_bitmap;
u8 io_bitmap_area[16384];

u8 *msr_bitmap;
u8 msr_bitmap_area[MSR_BITMAP_SIZE + PAGE_SIZE];

void vmcb_ident(struct vmcb *vmcb) {
  u64 vmcb_phys = virt_to_phys(vmcb);
  struct vmcb_save_area *save = &vmcb->save;
  struct vmcb_control_area *ctrl = &vmcb->control;
  u32 data_seg_attr = 3 | SVM_SELECTOR_S_MASK | SVM_SELECTOR_P_MASK |
                      SVM_SELECTOR_DB_MASK | SVM_SELECTOR_G_MASK;
  u32 code_seg_attr = 9 | SVM_SELECTOR_S_MASK | SVM_SELECTOR_P_MASK |
                      SVM_SELECTOR_L_MASK | SVM_SELECTOR_G_MASK;
  struct descriptor_table_ptr desc_table_ptr;

  memset(vmcb, 0, sizeof(*vmcb));
  asm volatile("vmsave %0" : : "a"(vmcb_phys) : "memory");
  vmcb_set_seg(&save->es, read_es(), 0, -1U, data_seg_attr);
  vmcb_set_seg(&save->cs, read_cs(), 0, -1U, code_seg_attr);
  vmcb_set_seg(&save->ss, read_ss(), 0, -1U, data_seg_attr);
  vmcb_set_seg(&save->ds, read_ds(), 0, -1U, data_seg_attr);
  sgdt(&desc_table_ptr);
  vmcb_set_seg(&save->gdtr, 0, desc_table_ptr.base, desc_table_ptr.limit, 0);
  sidt(&desc_table_ptr);
  vmcb_set_seg(&save->idtr, 0, desc_table_ptr.base, desc_table_ptr.limit, 0);
  ctrl->asid = 1;
  save->cpl = 0;
  save->efer = rdmsr(MSR_EFER);
  save->cr4 = read_cr4();
  save->cr3 = read_cr3();
  save->cr0 = read_cr0();
  save->dr7 = read_dr7();
  save->dr6 = read_dr6();
  save->cr2 = read_cr2();
  save->g_pat = rdmsr(MSR_IA32_CR_PAT);
  save->dbgctl = rdmsr(MSR_IA32_DEBUGCTLMSR);
  ctrl->intercept = (1ULL << INTERCEPT_VMRUN) | (1ULL << INTERCEPT_VMMCALL) |
                    (1ULL << INTERCEPT_MSR_PROT);
  ctrl->iopm_base_pa = virt_to_phys(io_bitmap);
  ctrl->msrpm_base_pa = virt_to_phys(msr_bitmap);
}

u64 guest_stack[10000];
u64 guest_stack2[10000];

struct regs regs;
struct regs regs2;

int _svm_vmrun(struct vmcb *vmcb, void *f) {
  vmcb->save.rip = (ulong)f;
  vmcb->save.rsp = (ulong)(guest_stack + ARRAY_SIZE(guest_stack));

  asm volatile(ASM_VMRUN_CMD(regs)
               :
               : "a"(virt_to_phys(vmcb))
               : "memory", "r15");

  return (vmcb->control.exit_code);
}

int nested_svm_vmrun(struct vmcb *vmcb, void *f) {
  vmcb->save.rip = (ulong)f;
  vmcb->save.rsp = (ulong)(guest_stack2 + ARRAY_SIZE(guest_stack2));

  asm volatile(ASM_VMRUN_CMD(regs2)
               :
               : "a"(virt_to_phys(vmcb))
               : "memory", "r15");

  return (vmcb->control.exit_code);
}

void setup_svm(void) {
  void *hsave = alloc_page();

  wrmsr(MSR_VM_HSAVE_PA, virt_to_phys(hsave));
  wrmsr(MSR_EFER, rdmsr(MSR_EFER) | EFER_SVME);
  wrmsr(MSR_EFER, rdmsr(MSR_EFER) | EFER_NX);

  io_bitmap = (void *)(((ulong)io_bitmap_area + 4095) & ~4095);

  msr_bitmap = (void *)ALIGN((ulong)msr_bitmap_area, PAGE_SIZE);
}

u64 scatter_bits(u64 qty, u64 fmt) {
  u64 reg = 0;
  u64 qty_pos = 1;
  u64 fmt_pos = 1;
  uint8_t i;
  for (i = 0; i < 64; i++) {
    if (fmt & fmt_pos) {
      if (qty & qty_pos) {
        reg |= fmt_pos;
      }
      qty_pos <<= 1;
    }
    fmt_pos <<= 1;
  }
  return reg;
}
