
#include "alloc.h"
#include "alloc_page.h"
#include "apic.h"
#include "delay.h"
#include "desc.h"
#include "isr.h"
#include "libcflat.h"
#include "msr.h"
#include "processor.h"
#include "smp.h"
#include "svm.h"
#include "vm.h"

#define SAVE_GPR_C(r)                                                          \
  "xchg %%rbx, " #r "+0x8\n\t"                                                 \
  "xchg %%rcx, " #r "+0x10\n\t"                                                \
  "xchg %%rdx, " #r "+0x18\n\t"                                                \
  "xchg %%rbp, " #r "+0x28\n\t"                                                \
  "xchg %%rsi, " #r "+0x30\n\t"                                                \
  "xchg %%rdi, " #r "+0x38\n\t"                                                \
  "xchg %%r8, " #r "+0x40\n\t"                                                 \
  "xchg %%r9, " #r "+0x48\n\t"                                                 \
  "xchg %%r10, " #r "+0x50\n\t"                                                \
  "xchg %%r11, " #r "+0x58\n\t"                                                \
  "xchg %%r12, " #r "+0x60\n\t"                                                \
  "xchg %%r13, " #r "+0x68\n\t"                                                \
  "xchg %%r14, " #r "+0x70\n\t"                                                \
  "xchg %%r15, " #r "+0x78\n\t"

#define LOAD_GPR_C SAVE_GPR_C

#define ASM_VMRUN_CMD(regs)                                                    \
  "vmload %%rax\n\t"                                                           \
  "mov " #regs "+0x80, %%r15\n\t"                                              \
  "mov %%r15, 0x170(%%rax)\n\t"                                                \
  "mov " #regs ", %%r15\n\t"                                                   \
  "mov %%r15, 0x1f8(%%rax)\n\t" LOAD_GPR_C(regs) "vmrun %%rax\n\t" SAVE_GPR_C( \
      regs) "mov 0x170(%%rax), %%r15\n\t"                                      \
            "mov %%r15, " #regs "+0x80\n\t"                                    \
            "mov 0x1f8(%%rax), %%r15\n\t"                                      \
            "mov %%r15, " #regs "\n\t"                                         \
            "vmsave %%rax\n\t"

// MSR names/identifiers are from https://github.com/jlgreathouse/AMD_IBS_Toolkit
#define MSR_IBS_OP_CTL 0xc0011033

#define MSR_IBS_OP_DATA3 0xc0011037
#define IBS_DC_PHY_ADDR_VALID (1ULL << 18)
#define IBS_DC_LIN_ADDR_VALID (1ULL << 17)
#define IBS_ST_OP (1ULL << 1)
#define IBS_LD_OP 1ULL

#define IBS_OP_VAL (1ULL << 18)
#define IBS_OP_CUR_CNT_23 (0x7fffff0ULL << 32)
#define IBS_OP_MAX_CNT (0x7f0ffffULL) /* LN+ */
#define IBS_OP_CNT_CTL (1ULL << 19)
#define IBS_OP_EN (1ULL << 17)
#define MSR_IBS_OP_DATA 0xc0011035
#define MSR_IBS_OP_RIP 0xc0011034

#define MSR_IBS_DC_LIN_AD 0xc0011038
#define MSR_IBS_DC_PHYS_AD 0xc0011039

#define MSR_IBS_FETCH_CTL 0xc0011030
#define IBS_RAND_EN (1ULL << 57)
#define IBS_L2_TLB_MISS (1ULL << 56)
#define IBS_L1_TLB_MISS (1ULL << 55)
#define IBS_L1_TLB_PG_SZ (3ULL << 53)
#define IBS_PHY_ADDR_VALID (1ULL << 52)
#define IBS_IC_MISS (1ULL << 51)
#define IBS_FETCH_COMP (1ULL << 50)
#define IBS_FETCH_VAL (1ULL << 49)
#define IBS_FETCH_EN (1ULL << 48)
#define IBS_FETCH_LAT (0xffffULL << 32)
#define IBS_FETCH_CNT (0xffffULL << 16)
#define IBS_FETCH_MAX_CNT 0xffffULL

#define MSR_IBS_FETCH_LIN_AD 0xc0011031
#define MSR_IBS_FETCH_PHYS_AD 0xc0011032

#define MSR_KVM_SYSTEM_TIME_NEW 0x4b564d01

u64 scatter_bits(u64 qty, u64 fmt);
void setup_svm(void);
int nested_svm_vmrun(struct vmcb *vmcb, void *f);
int _svm_vmrun(struct vmcb *vmcb, void *f);
void vmcb_ident(struct vmcb *vmcb);
void vmcb_set_seg(struct vmcb_seg *seg, u16 selector, u64 base, u32 limit,
                  u32 attr);
